<?php

namespace Drupal\Tests\ggroup\Kernel;

use Drupal\group\Entity\GroupContent;
use Drupal\Tests\group\Kernel\GroupKernelTestBase;

/**
 * Tests the behavior of subgroup creators.
 *
 * @group group
 */
class SubgroupTest extends GroupKernelTestBase {

  /**
   * The group hierarchy manager.
   *
   * @var \Drupal\ggroup\GroupHierarchyManager
   */
  protected $groupHierarchyManager;

  /**
   * The account to use as the group creator.
   *
   * @var \Drupal\group\Entity\GroupTypeInterface
   */
  protected $groupType;

  /**
   * The account to use as the group creator.
   *
   * @var \Drupal\group\Entity\GroupTypeInterface
   */
  protected $subGroupType;

  /**
   * Modules to enable.
   *
   * @var array
   */
  protected static $modules = [
    'ggroup',
    'ggroup_test_config',
  ];

  /**
   * {@inheritdoc}
   */
  protected function setUp(): void {
    parent::setUp();

    $this->installConfig(['ggroup_test_config']);
    $this->installSchema('ggroup', 'group_graph');

    $this->entityTypeManager = $this->container->get('entity_type.manager');
    $this->groupHierarchyManager = $this->container->get('ggroup.group_hierarchy_manager');

    $this->groupType = $this->entityTypeManager->getStorage('group_type')->load('default');
    $this->subGroupType = $this->entityTypeManager->getStorage('group_type')->load('subgroup');
  }

  /**
   * Tests the addition of a group to a group.
   */
  public function testCreateSubgroup() {
    [$group, $sub_group] = $this->addGroup();
    $this->assertNotEmpty($group->getContentByEntityId('subgroup:' . $this->subGroupType->id(), $sub_group->id()), 'Subgroup is group content');
  }

  /**
   * Tests the removing subgroup from group.
   */
  public function testDeleteSubgroupFromGroupContent() {
    /** @var \Drupal\group\Entity\Group $sub_group */
    [$group, $sub_group] = $this->addGroup();

    foreach (GroupContent::loadByEntity($sub_group) as $group_content) {
      $group_content->delete();

      $this->assertEquals($this->groupHierarchyManager->groupHasSubgroup($group, $sub_group), FALSE, 'Subgroup is removed');
    }

  }

  /**
   * Tests the removing subgroup.
   */
  public function testDeleteSubgroup() {
    [$group, $sub_group] = $this->addGroup();

    /** @var \Drupal\group\Entity\Group $sub_group */
    $sub_group->delete();

    $this->assertEquals($this->groupHierarchyManager->groupHasSubgroup($group, $sub_group), FALSE, 'Subgroup is removed');
  }

  /**
   * Tests get subgroups.
   */
  public function testGetGroupSubgroups() {
    $subgroup_ids = [];
    [$group, $sub_group] = $this->addGroup();
    $subgroup_ids[] = $sub_group->id();
    /** @var \Drupal\group\Entity\Group $sub_group2 */
    $sub_group2 = $this->createGroupByType($this->subGroupType->id(), ['uid' => $this->getCurrentUser()->id()]);
    $subgroup_ids[] = $sub_group2->id();
    $group->addContent($sub_group2, 'subgroup:' . $this->subGroupType->id());

    $sub_groups = $this->groupHierarchyManager->getGroupSubgroups($group->id());
    $this->assertEquals(2, count($sub_groups), 'Group has 2 subgroups');

    foreach ($sub_groups as $sub_group_item) {
      $this->assertEquals(TRUE, in_array($sub_group_item->id(), $subgroup_ids), 'Group is among subgroups');
    }
  }

  /**
   * Tests get subgroups ids.
   */
  public function testGetGroupSubgroupIds() {
    $subgroup_ids = [];
    [$group, $sub_group] = $this->addGroup();
    $subgroup_ids[] = $sub_group->id();
    $sub_group2 = $this->createGroupByType($this->subGroupType->id(), ['uid' => $this->getCurrentUser()->id()]);
    $subgroup_ids[] = $sub_group2->id();
    $group->addContent($sub_group2, 'subgroup:' . $this->subGroupType->id());

    $sub_group_ids = $this->groupHierarchyManager->getGroupSubgroupIds($group->id());
    $this->assertEquals(2, count($sub_group_ids), 'Group has 2 subgroups');

    foreach ($sub_group_ids as $sub_group_id) {
      $this->assertEquals(TRUE, in_array($sub_group_id, $subgroup_ids), 'Group is among subgroups');
    }
  }

  /**
   * Tests get group super groups.
   */
  public function testGetGroupSupergroups() {
    $super_group_ids = [];
    [$group, $sub_group] = $this->addGroup();
    $sub_group2 = $this->createGroupByType($this->subGroupType->id(), ['uid' => $this->getCurrentUser()->id()]);
    $super_group_ids[] = $group->id();
    $super_group_ids[] = $sub_group->id();
    $sub_group->addContent($sub_group2, 'subgroup:' . $this->subGroupType->id());

    $super_groups = $this->groupHierarchyManager->getGroupSupergroups($sub_group2->id());
    $this->assertEquals(2, count($super_groups), 'Group has 2 supergroups');

    foreach ($super_groups as $super_group) {
      $this->assertEquals(TRUE, in_array($super_group->id(), $super_group_ids), 'Group is among supergroups');
    }
  }

  /**
   * Tests get group super groups ids.
   */
  public function testGetGroupSupergroupIds() {
    $super_group_ids = [];
    [$group, $sub_group] = $this->addGroup();
    $sub_group2 = $this->createGroupByType($this->subGroupType->id(), ['uid' => $this->getCurrentUser()->id()]);
    $super_group_ids[] = $group->id();
    $super_group_ids[] = $sub_group->id();
    $sub_group->addContent($sub_group2, 'subgroup:' . $this->subGroupType->id());

    $super_groups = $this->groupHierarchyManager->getGroupSupergroupIds($sub_group2->id());
    $this->assertEquals(2, count($super_groups), 'Group has 2 supergroups');

    foreach ($super_groups as $super_group) {
      $this->assertEquals(TRUE, in_array($super_group, $super_group_ids), 'Group is among supergroups');
    }
  }

  /**
   * Create group and attach subgroup to group.
   *
   * @return array
   *   Return group and subgroup.
   */
  private function addGroup() {
    /** @var \Drupal\group\Entity\Group $group */
    $group = $this->createGroupByType($this->groupType->id(), ['uid' => $this->getCurrentUser()->id()]);
    /** @var \Drupal\group\Entity\Group $sub_group */
    $sub_group = $this->createGroupByType($this->subGroupType->id(), ['uid' => $this->getCurrentUser()->id()]);

    $group->addContent($sub_group, 'subgroup:' . $this->subGroupType->id());

    return [$group, $sub_group];
  }

  /**
   * Creates a group by type.
   *
   * @param string $type
   *   Group type.
   * @param array $values
   *   (optional) The values used to create the entity.
   *
   * @return \Drupal\group\Entity\Group
   *   The created group entity.
   */
  private function createGroupByType($type, array $values = []) {
    /** @var \Drupal\group\Entity\Group $group */
    $group = $this->entityTypeManager->getStorage('group')->create($values + [
      'type' => $type,
      'label' => $this->randomMachineName(),
    ]);

    $group->enforceIsNew();
    $group->save();

    return $group;
  }

}
